﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using System.Net;
using System.Net.Sockets;
using System.Threading;



namespace Traceroute
{
    public partial class TracerouteMain : Form
    {
        BindingList<Hop> mHops = new BindingList<Hop>();
        private Thread mDGVWorker;
        private bool mStopThread;
        private string mHostToTrace;


        public TracerouteMain()
        {
            mStopThread = false;
            InitializeComponent();
        }




        private void TracerouteMain_Load(object sender, EventArgs e)
        {

        }





        private void CloseButton_Click(object sender, EventArgs e)
        {
            stopButton();
            Dispose();
        }






        private void StartButton_Click(object sender, EventArgs e)
        {
            IPHostEntry lHostEntry;
            Regex lRegex;
            Match lMatch;







            /*
             * 
             * 
             */

            if (this.StartButton.Text == "Stop")
            {
                stopButton();
            }
            else
            {

                /*
                 * Set buttons into a neutral state
                 * 
                 */

                this.StartButton.Text = "Stop";
                this.ClearButton.Enabled = false;
                this.StartButton.Enabled = false;
                this.CloseButton.Enabled = false;
                this.HostTextBox.Enabled = false;
                this.Cursor = Cursors.WaitCursor;



                /*
                 * Make sure no other thread is running
                 * 
                 */

                if (mDGVWorker != null && mDGVWorker.IsAlive)
                {
                    this.mStopThread = true;
                    mDGVWorker.Abort();
                    mDGVWorker.Join();
                }
                this.mStopThread = false;


                /*
                 * Resolve hostname and start trace
                 * 
                 */

                mHops.Clear();
                RouteDataGridView.DataSource = mHops;
                RouteDataGridView.Refresh();

                mHostToTrace = this.HostTextBox.Text.ToString();


                if (this.HostTextBox.Text.ToString().Length <= 0)
                {
                    MessageBox.Show("No host name defined.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    stopButton();
                }
                else
                {
                    lRegex = new Regex(@"^\d+\.\d+\.\d+\.\d+$");
                    lMatch = lRegex.Match(this.HostTextBox.Text.ToString());

                    if (lMatch.Success)
                    {
                        HostLabel.Text = this.HostTextBox.Text;
                        mDGVWorker = new Thread(new ThreadStart(startTrace));
                        mDGVWorker.IsBackground = true;
                        mDGVWorker.Start();
                    }
                    else
                    {
                        try
                        {
                            lHostEntry = Dns.Resolve(this.HostTextBox.Text.ToString());
                            this.StartButton.Enabled = true;

                            if (lHostEntry.AddressList[0].ToString().Length > 0)
                            {
                                HostLabel.Text = lHostEntry.AddressList[0].ToString();
                                mDGVWorker = new Thread(new ThreadStart(startTrace));
                                mDGVWorker.IsBackground = true;
                                mDGVWorker.Start();
                            }
                            else
                            {
                                displayMessage("Something is wrong with the host name.");
                                stopButton();
                            }
                        }
                        catch (Exception ex)
                        {
                            displayMessage("Something is wrong with the host name.");
                            stopButton();
                        }
                    }
                }
            }
        }




        private void startTrace()
        {
            ushort lCounter;
            int lPacketSize = 1024;
            int lTTL;
            int lMaxHops = 30;
            bool lTraceOk = false;
            bool lReachable = false;
            int lErrorCounter = 0;
            IPPacket lIP;
            ICMPSocket lSocket = new ICMPSocket();
            ICMPPacket lICMP = new ICMPPacket();
            ICMPEcho lMessage = new ICMPEcho();





            /*
             * Check if remote system is reachable.
             * 
             */

            System.Random Random = new System.Random();
            lICMP.Message = lMessage;


            lMessage.Identifier = (ushort)Random.Next(0, 0xffff);
            lMessage.SequenceNumber = 1;
            lMessage.Data = "www.megapanzer.com".PadRight(lPacketSize, '*');

            try
            {
                lIP = lSocket.Send(mHostToTrace, lICMP, 128);


                    if (lIP.ICMP.Message is ICMPEchoReply)
                    {
                        lReachable = true;
                    }
                    else if (lIP.ICMP.Message is ICMPTimeExceeded)
                    {
                        lReachable = false;
                        displayMessage("Time Exceeded.");
                    }
                    else if (lIP.ICMP.Message is ICMPDestinationUnreachable)
                    {
                        lReachable = false;
                        displayMessage("Destination Unreachable.");
                    }
                    else
                    {
                        lReachable = false;
                        if (mStopThread == false)
                            MessageBox.Show(lIP.ICMP.Message.ToString(), "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
                 
            }
            catch (Exception e)
            {
                lReachable = false;
                displayMessage("The following error occurred : " + e.ToString() + "\r\n\r\n");
            }





            
            /*
             * Trace remote system
             * 
             */

            if (lReachable == true && mStopThread == false)
            {
                for (lTTL = 1; lTTL < lMaxHops + 1 && mStopThread == false; lTTL++)
                {
                    lICMP = new ICMPPacket();
                    lMessage = new ICMPEcho();
                    lICMP.Message = lMessage;


                    for (lCounter = 1; lCounter < 4 && mStopThread == false; lCounter++)
                    {
                        lMessage.Identifier = 1;
                        lMessage.SequenceNumber = lCounter;
                        lMessage.Data = "www.megapanzer.com".PadRight(lPacketSize, '*');

                        try
                        {
                            IPPacket IP = lSocket.Send(mHostToTrace, lICMP, lTTL);
                            string ResponseAddress = ((System.Net.IPEndPoint)lSocket.EndPoint).Address.ToString();



                            /*
                             * Add new item
                             * 
                             */

                            if (IP.ICMP.Message is ICMPTimeExceeded)
                            {
                                updateDelegate(ResponseAddress, Dns.Resolve(ResponseAddress).HostName, lSocket.TimeElapsed.ToString() + " ms      ");
                                break;
                            }
                            else if (IP.ICMP.Message is ICMPEchoReply)
                            {
                                updateDelegate(ResponseAddress, Dns.Resolve(ResponseAddress).HostName, lSocket.TimeElapsed.ToString() + " ms      ");
                                lTraceOk = true;
                                break;
                            }
                            else if (IP.ICMP.Message is ICMPIPHeaderReply)
                            {
                                lTraceOk = true;
                                break;
                            }
                            lErrorCounter = 0;
                        }
                        catch
                        {
                            lErrorCounter++;
                        }
                    }

                    if (lTraceOk || lErrorCounter > 8)
                        break;
                }
            }
            this.Invoke((MethodInvoker)delegate() { stopButton(); });
        }




        private void displayMessage(string pMessage)
        {
            if (mStopThread == false)
                MessageBox.Show(pMessage, "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }





        private void updateDelegate(string pIP, string pHostName, string pTime)
        {
            if (this.InvokeRequired == true)
                this.Invoke((MethodInvoker)delegate() { addHop(pIP, pHostName, pTime); });
            else
                addHop(pIP, pHostName, pTime);
        }







        private void addHop(string pIP, string pHostName, string pTime)
        {
            
            if (mStopThread == false)
            {
                mHops.Add(new Hop(pIP, pHostName, pTime));
                RouteDataGridView.DataSource = mHops;
                RouteDataGridView.Refresh();
            }
        }

        


        private void ClearButton_Click(object sender, EventArgs e)
        {
            HostLabel.Text = "";
            mHops.Clear();
            RouteDataGridView.DataSource = mHops;
            RouteDataGridView.Refresh();
        }





        private void stopButton()
        {
            this.StartButton.Text = "Start";
            this.ClearButton.Enabled = true;
            this.StartButton.Enabled = true;
            this.CloseButton.Enabled = true;
            this.HostTextBox.Enabled = true;
            this.Cursor = Cursors.Default;
            mStopThread = true;
        }
    }
}
